"""Embeddable Acherion workbench wrapper around the shared designer."""

from __future__ import annotations

import json
from collections.abc import Mapping
from contextlib import nullcontext
from typing import Any, Callable

from nicegui import background_tasks, ui

import acherion.preferences as acherion_preferences
from acherion.embed.code_editor import build_python_code_editor
from acherion.embed.designer.component import AcherionDesigner
from acherion.host import AcherionHost
from acherion.theme import (
    build_scoped_acherion_theme_css,
    normalize_acherion_theme_overrides,
)


class AcherionWorkbench:
    """Embeddable and customizable workbench shell for AcherionDesigner.

    Hosts can use the default generated-code pane and validation behavior, or
    inject their own code view and callbacks while still reusing the same
    workbench component structure across applications.
    """

    def __init__(
        self,
        *,
        host: AcherionHost | None = None,
        on_change: Callable[[], None] | None = None,
        on_local_change: Callable[[], None] | None = None,
        on_apply_to_code: Callable[[], None] | None = None,
        on_run_preview: Callable[[], bool] | None = None,
        on_validate: Callable[[], bool] | None = None,
        build_code_view: Callable[[], Any] | None = None,
        on_mode_change: Callable[[str], None] | None = None,
        initial_mode: str = 'graph',
        before_build: Callable[[], None] | None = None,
        container_classes: str = '',
        code_editor_theme: Any = 'vscodeDark',
        code_editor_classes: str = 'w-full h-full oe-code-editor',
        code_editor_style: str = 'height:100%; min-height:0',
        code_transform: Callable[[str], str] | None = None,
        refresh_code_on_change: bool = False,
        refresh_code_after_build: bool = False,
        validate_generated_code: Callable[[str], Any] | None = None,
        apply_to_code_status_message: str | None = None,
        theme_overrides: Mapping[str, str] | None = None,
        preferences: Mapping[str, Any] | None = None,
        load_preferences: Callable[[], Mapping[str, Any] | None] | None = None,
        save_preferences: (
            Callable[[dict[str, dict[str, str]]], None] | None
        ) = None,
        session_storage_key: str | None = None,
    ) -> None:
        self._on_change = on_change
        self._on_local_change = on_local_change
        self._on_apply_to_code = on_apply_to_code
        self._on_validate = on_validate
        self._host_build_code_view = build_code_view
        self._before_build = before_build
        self._container_classes = container_classes
        self._code_editor_classes = code_editor_classes
        self._code_editor_style = code_editor_style
        self._code_transform = code_transform
        self._refresh_code_on_change = refresh_code_on_change
        self._refresh_code_after_build = refresh_code_after_build
        self._validate_generated_code = validate_generated_code
        self._apply_to_code_status_message = apply_to_code_status_message
        initial_preferences = {
            'appearance': {
                'code_editor_theme': str(code_editor_theme or '').strip()
                or acherion_preferences.DEFAULT_CODE_EDITOR_THEME,
            },
        }
        self._preferences_state = (
            acherion_preferences.AcherionPreferencesState(initial_preferences)
        )
        self._preferences_state.apply_mapping(preferences)
        if load_preferences is not None:
            self._preferences_state.apply_mapping(load_preferences())
        self._code_editor_theme = self._preferences_state.code_editor_theme
        self._save_preferences = save_preferences
        self._session_storage_key = str(session_storage_key or '').strip()
        self._theme_overrides = (
            dict(theme_overrides)
            if theme_overrides is not None
            else None
        )
        self._theme_css_injected = False
        self._built = False
        self._applied_theme_variable_names: set[str] = set()
        self._code_view: Any = None
        self._designer = AcherionDesigner(
            host=host,
            on_change=self._handle_change,
            on_local_change=self._handle_local_change,
            on_apply_to_code=self._handle_apply_to_code,
            on_run_preview=on_run_preview,
            on_validate=self._handle_validate,
            build_code_view=self._build_code_view,
            on_mode_change=on_mode_change,
            preferences_state=self._preferences_state,
            on_preferences_change=self._handle_preferences_change,
            on_preferences_preview=self._preview_preferences,
            initial_mode=initial_mode,
        )

    @property
    def designer(self) -> AcherionDesigner:
        """Return the underlying AcherionDesigner instance."""
        return self._designer

    @property
    def code_view(self) -> Any:
        """Return the registered code-view widget, if one exists."""
        return self._code_view

    def preferences_dict(self) -> dict[str, dict[str, str]]:
        """Return the current workbench preferences as a plain dictionary."""
        return self._preferences_state.to_dict()

    def set_code_view(self, code_view: Any) -> Any:
        """Register the current code-view widget for refresh operations."""
        self._code_view = code_view
        self._apply_code_editor_theme()
        return code_view

    def set_preferences(
        self,
        preferences: Mapping[str, Any] | None,
        *,
        persist: bool = True,
    ) -> None:
        """Replace current preferences and apply the new runtime state."""
        if not self._preferences_state.apply_mapping(preferences):
            if persist:
                self._persist_preferences()
            return
        self._apply_preference_state(persist=persist)

    def set_theme_overrides(
        self,
        theme_overrides: Mapping[str, str] | None,
    ) -> None:
        """Enable, replace, or disable the embedded scoped Acherion theme."""
        self._theme_overrides = (
            dict(theme_overrides)
            if theme_overrides is not None
            else None
        )
        if self._theme_overrides is not None:
            self._ensure_theme_css()
        self._apply_theme_state()

    def clear_theme_overrides(self) -> None:
        """Disable the embedded scoped Acherion theme for this workbench."""
        self.set_theme_overrides(None)

    def build(self) -> None:
        """Render the workbench into the current NiceGUI context."""
        if self._before_build is not None:
            self._before_build()
        if self._theme_overrides is not None:
            self._ensure_theme_css()
        container = (
            ui.element('div').classes(self._container_classes)
            if self._container_classes.strip()
            else nullcontext()
        )
        with container:
            self._designer.build()
        self._built = True
        self._apply_theme_state()
        self._apply_preference_state(persist=False)
        if self._refresh_code_after_build:
            self.refresh_code_view()
        self._load_session_preferences()

    def generated_user_code(self) -> str:
        """Return the underlying designer's generated user code."""
        return self._designer.generated_user_code()

    def refresh_code_view(self) -> None:
        """Refresh the registered code-view widget from generated code."""
        if self._code_view is None:
            return
        self._set_code_view_value(
            self._transform_display_code(self.generated_user_code())
        )

    def _transform_display_code(self, code: str) -> str:
        """Apply optional host-side display transforms to generated code."""
        if self._code_transform is None:
            return code
        return str(self._code_transform(code))

    def _theme_scope_selector(self) -> str:
        """Return the scoped selector used for the embedded Acherion theme."""
        return f'#{self._designer.frame_dom_id}[data-ach-theme="enabled"]'

    def _ensure_theme_css(self) -> None:
        """Inject the scoped embedded theme stylesheet once."""
        if self._theme_css_injected:
            return
        ui.add_css(build_scoped_acherion_theme_css(self._theme_scope_selector()))
        self._theme_css_injected = True

    def _apply_theme_state(self) -> None:
        """Apply the current runtime theme state to an already-built workbench."""
        if not self._built:
            return
        next_variables = normalize_acherion_theme_overrides(self._theme_overrides)
        remove_names = sorted(
            self._applied_theme_variable_names - set(next_variables)
            if self._theme_overrides is not None
            else self._applied_theme_variable_names
        )
        payload = json.dumps({
            'frameId': self._designer.frame_dom_id,
            'enabled': self._theme_overrides is not None,
            'variables': next_variables,
            'removeNames': remove_names,
        })
        self._designer.run_client_javascript(
            '(() => {'
            f'const payload = {payload};'
            'const frame = document.getElementById(payload.frameId);'
            'if (!frame) return;'
            'if (payload.enabled) {'
            '  frame.setAttribute("data-ach-theme", "enabled");'
            '} else {'
            '  frame.removeAttribute("data-ach-theme");'
            '}'
            'for (const name of payload.removeNames) {'
            '  frame.style.removeProperty(name);'
            '}'
            'for (const [name, value] of Object.entries(payload.variables)) {'
            '  frame.style.setProperty(name, String(value));'
            '}'
            '})();'
        )
        self._applied_theme_variable_names = set(next_variables)

    def _set_code_view_value(self, value: str) -> None:
        """Update the registered code-view widget with one text value."""
        if self._code_view is None:
            return
        setter = getattr(self._code_view, 'set_value', None)
        if callable(setter):
            setter(value)
            return
        if hasattr(self._code_view, 'value'):
            self._code_view.value = value

    def _build_code_view(self) -> None:
        """Build either the host-provided or default generated-code view."""
        if self._host_build_code_view is not None:
            code_view = self._host_build_code_view()
            if code_view is not None:
                self.set_code_view(code_view)
            return
        with ui.element('div').classes(
            'w-full h-full min-w-0 min-h-0 flex flex-1'
        ):
            editor = build_python_code_editor(
                value='',
                theme=self._preferences_state.code_editor_theme,
                classes=self._code_editor_classes,
                style=self._code_editor_style,
            )
        self.set_code_view(editor)

    def _apply_code_editor_theme(self) -> None:
        """Push the active editor theme into the current code view."""
        if self._code_view is None:
            return
        theme_name = self._preferences_state.code_editor_theme
        setter = getattr(self._code_view, 'set_theme', None)
        if callable(setter):
            setter(theme_name)
            return
        if hasattr(self._code_view, 'theme'):
            self._code_view.theme = theme_name

    def _apply_preference_state(self, *, persist: bool) -> None:
        """Apply shared preference state to runtime widgets and hooks."""
        self._code_editor_theme = self._preferences_state.code_editor_theme
        self._apply_code_editor_theme()
        if self._built:
            apply_preferences = getattr(
                self._designer,
                'apply_preferences_state',
                None,
            )
            if callable(apply_preferences):
                apply_preferences()
        if persist:
            self._persist_preferences()

    def _handle_preferences_change(
        self,
        preferences: dict[str, dict[str, str]],
    ) -> None:
        """Persist designer-side preference edits through the wrapper."""
        self.set_preferences(preferences)

    def _preview_preferences(
        self,
        preferences: dict[str, dict[str, str]],
    ) -> None:
        """Apply non-persisted preference previews from the dialog."""
        self.set_preferences(preferences, persist=False)

    def _persist_preferences(self) -> None:
        """Persist the current preferences through configured backends."""
        preferences_payload = self.preferences_dict()
        if self._save_preferences is not None:
            self._save_preferences(preferences_payload)
        if self._session_storage_key:
            payload = json.dumps(preferences_payload)
            self._designer.set_session_storage_item(
                self._session_storage_key,
                payload,
            )

    def _load_session_preferences(self) -> None:
        """Load standalone preferences from session storage after build."""
        if not self._session_storage_key:
            return
        async def _load() -> None:
            raw_preferences = await self._designer.get_session_storage_item(
                self._session_storage_key
            )
            if not raw_preferences:
                return
            try:
                loaded_preferences = json.loads(str(raw_preferences))
            except json.JSONDecodeError:
                return
            self.set_preferences(loaded_preferences, persist=False)

        background_tasks.create(
            _load(),
            name='acherion load session preferences',
        )

    def _handle_change(self) -> None:
        """Dispatch designer change notifications through the wrapper."""
        if self._refresh_code_on_change:
            self.refresh_code_view()
        if self._on_change is not None:
            self._on_change()

    def _handle_local_change(self) -> None:
        """Dispatch lightweight inline-change notifications."""
        if self._on_local_change is None:
            self._handle_change()
            return
        if self._refresh_code_on_change:
            self.refresh_code_view()
        self._on_local_change()

    def _handle_apply_to_code(self) -> None:
        """Dispatch compile/apply requests or fall back to code refresh."""
        if self._on_apply_to_code is not None:
            self._on_apply_to_code()
            return
        self.refresh_code_view()
        if self._apply_to_code_status_message:
            self._designer.set_status_message(
                self._apply_to_code_status_message
            )

    def _handle_validate(self) -> bool:
        """Validate generated code through host callback or default handler."""
        if self._on_validate is not None:
            return self._on_validate()
        if self._validate_generated_code is None:
            self._designer.set_status_message(
                'Generated code validation is not available.',
                negative=True,
            )
            return False
        try:
            self._validate_generated_code(self.generated_user_code())
        except Exception as exc:  # Validation boundary for host feedback.
            self.refresh_code_view()
            self._designer.set_status_message(
                f'Generated code validation failed: {exc}',
                negative=True,
            )
            return False
        self.refresh_code_view()
        self._designer.set_status_message('Generated code validated successfully.')
        return True


__all__ = ['AcherionWorkbench']


"""Pin and node-card rendering mixin for AcherionDesigner."""

# pyright: reportAttributeAccessIssue=false

from __future__ import annotations

from typing import Any, cast

from nicegui import ui

from acherion.catalog import types as _catalog_types
import acherion.node_behaviors as acherion_node_behaviors
from acherion.model import (
    AcherionNode,
    _node_var_name,
    _template_icon,
    _template_title,
)


def _pin_button_classes(base: str, pin_type: str) -> str:
    """Return CSS classes for a typed pin button."""
    style_tag = _catalog_types.pin_style_tag(pin_type)
    return f'{base} ach-pin-btn-type-{style_tag}'


def _type_badge_classes(pin_type: str) -> str:
    """Return CSS classes for a typed badge label."""
    style_tag = _catalog_types.pin_style_tag(pin_type)
    return f'ach-type-badge ach-type-badge-{style_tag}'


def _is_exec_pin(pin: dict[str, str]) -> bool:
    """Return True when a pin spec represents control flow."""
    return str(pin.get('type') or '') == 'exec'


def _is_unlabeled_exec_pin(pin: dict[str, str]) -> bool:
    """Return True for exec pins that should live on the top row."""
    if not _is_exec_pin(pin):
        return False
    return not str(pin.get('label') or '').strip()


def _event_bool_arg(args: Any, key: str) -> bool:
    """Return bool event arg when present."""
    if not isinstance(args, dict):
        return False
    value = cast(dict[str, object], args).get(key)
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        return value.strip().lower() in {'1', 'true', 'yes', 'on'}
    return bool(value)


def _pairable_body_pin_key(
    pin: dict[str, str],
) -> tuple[str, str] | None:
    """Return stable match key for inline bidirectional data rows."""
    label = str(pin.get('label') or '').strip()
    pin_type = str(pin.get('type') or 'any').strip() or 'any'
    if not label or pin_type == 'exec':
        return None
    return (label, pin_type)


class _RenderPinsMixin:
    """Pin-row and system-node rendering methods."""

    @staticmethod
    def _inline_default_editor_spec(
        node: AcherionNode,
    ) -> tuple[str, str, Any] | None:
        """Return inline default editor binding for supported nodes."""
        return acherion_node_behaviors.inline_default_editor_spec_for_node(node)

    def _set_inline_default_value(
        self: Any,
        node: AcherionNode,
        *,
        field_name: str,
        input_kind: str,
        value: Any,
        notify: bool = True,
        refresh: bool = True,
    ) -> None:
        """Persist one inline node default value."""
        if input_kind == 'number':
            if node.kind == 'constant':
                value_type = str(node.params.get('value_type') or 'number').strip()
                if value_type == 'int':
                    try:
                        node.params[field_name] = int(float(value or 0))
                    except (TypeError, ValueError):
                        node.params[field_name] = 0
                else:
                    node.params[field_name] = 0 if value in (None, '') else value
            else:
                node.params[field_name] = 0 if value in (None, '') else value
        else:
            node.params[field_name] = str(value or '')
        if notify:
            if refresh:
                self._notify_change()
            else:
                self._notify_change_without_refresh()

    def _mark_pending_inline_local_change(self: Any) -> None:
        self._pending_inline_local_change = True

    def _flush_pending_inline_local_change(
        self: Any,
        event: Any,
    ) -> None:
        if _event_bool_arg(getattr(event, 'args', None), 'keep_pending'):
            return
        if not bool(getattr(self, '_pending_inline_local_change', False)):
            return
        self._notify_change_without_refresh()

    def _render_inline_default_editor(self: Any, node: AcherionNode) -> bool:
        """Render compact node-card editor for supported default values."""
        if acherion_node_behaviors.render_inline_controls_for_node(self, node):
            return True
        spec = self._inline_default_editor_spec(node)
        if spec is None:
            return False
        input_kind, field_name, current_value = spec
        field_classes = 'ach-node-inline-field'
        if input_kind == 'number':
            field_classes += ' ach-node-inline-field-number'
            step = '1' if (
                node.kind == 'constant'
                and str(node.params.get('value_type') or 'number').strip() == 'int'
            ) else 'any'
            number_field = ui.number(
                value=current_value,
            ).props(
                f'dense outlined hide-bottom-space step={step}'
            ).classes(field_classes)

            def _stage_number_value(event: Any) -> None:
                self._set_inline_default_value(
                    node,
                    field_name=field_name,
                    input_kind=input_kind,
                    value=getattr(event, 'args', None),
                    notify=False,
                )
                self._mark_pending_inline_local_change()

            def _commit_number_value(event: Any) -> None:
                self._set_inline_default_value(
                    node,
                    field_name=field_name,
                    input_kind=input_kind,
                    value=number_field.value,
                    notify=False,
                )
                self._flush_pending_inline_local_change(event)

            number_field.on(
                'update:model-value',
                _stage_number_value,
            )
            number_field.on(
                'blur',
                _commit_number_value,
                js_handler=(
                    '(e) => emit({'
                    'keep_pending: !!('
                    'e.relatedTarget && '
                    'e.relatedTarget.closest(".ach-node-inline-field")'
                    ')'
                    '})'
                ),
            )
            number_field.on('keydown.enter', _commit_number_value)
            return True

        text_field = ui.input(
            value=str(current_value or ''),
        ).props('dense outlined hide-bottom-space').classes(
            field_classes + ' ach-node-inline-field-text'
        )

        def _stage_text_value(event: Any) -> None:
            self._set_inline_default_value(
                node,
                field_name=field_name,
                input_kind=input_kind,
                value=getattr(event, 'args', None),
                notify=False,
            )
            self._mark_pending_inline_local_change()

        def _commit_text_value(event: Any) -> None:
            self._set_inline_default_value(
                node,
                field_name=field_name,
                input_kind=input_kind,
                value=text_field.value,
                notify=False,
            )
            self._flush_pending_inline_local_change(event)

        text_field.on('update:model-value', _stage_text_value)
        text_field.on(
            'blur',
            _commit_text_value,
            js_handler=(
                '(e) => emit({'
                'keep_pending: !!('
                'e.relatedTarget && '
                'e.relatedTarget.closest(".ach-node-inline-field")'
                ')'
                '})'
            ),
        )
        text_field.on('keydown.enter', _commit_text_value)
        return True

    @staticmethod
    def _pin_literal_input_kind(
        pin_type: str,
        editor_kind: str = '',
    ) -> str | None:
        """Return compact editor kind for one eligible input pin type."""
        if editor_kind in {'number', 'text'}:
            return editor_kind
        if pin_type in {'float', 'int'}:
            return 'number'
        if pin_type == 'str':
            return 'text'
        return None

    @staticmethod
    def _pin_literal_value(
        node: AcherionNode,
        pin_id: str,
    ) -> Any:
        """Return stored literal fallback value for one input pin."""
        raw_literals = node.params.get('pin_literals')
        if not isinstance(raw_literals, dict):
            return None
        return raw_literals.get(pin_id)

    @staticmethod
    def _coerce_pin_literal_value(
        value: Any,
        *,
        input_kind: str,
        pin_type: str,
    ) -> Any:
        """Normalize one literal input value using the pin's declared type."""
        if input_kind != 'number':
            return str(value)
        if pin_type == 'int':
            try:
                return int(float(value))
            except (TypeError, ValueError):
                return None
        if pin_type == 'float':
            try:
                return float(value)
            except (TypeError, ValueError):
                return None
        return value

    def _set_pin_literal_value(
        self: Any,
        node: AcherionNode,
        *,
        pin_id: str,
        input_kind: str,
        pin_type: str,
        value: Any,
        notify: bool = True,
        refresh: bool = True,
    ) -> None:
        """Persist one literal fallback value for an input pin."""
        literals = dict(node.params.get('pin_literals') or {})
        coerced_value = self._coerce_pin_literal_value(
            value,
            input_kind=input_kind,
            pin_type=pin_type,
        )
        if coerced_value in (None, ''):
            literals.pop(pin_id, None)
        else:
            literals[pin_id] = coerced_value
        node.params['pin_literals'] = literals
        if notify:
            if refresh:
                self._notify_change()
            else:
                self._notify_change_without_refresh()

    def _render_input_literal_editor(
        self: Any,
        node: AcherionNode,
        *,
        pin_id: str,
        pin_type: str,
        editor_kind: str,
        source_id: str,
    ) -> bool:
        """Render compact literal editor for one unconnected input pin."""
        if source_id:
            return False
        input_kind = self._pin_literal_input_kind(pin_type, editor_kind)
        if input_kind is None:
            return False
        current_value = self._pin_literal_value(node, pin_id)
        field_classes = 'ach-node-inline-field'
        if input_kind == 'number':
            field_classes += ' ach-node-inline-field-number'
            step = '1' if pin_type == 'int' else 'any'
            number_field = ui.number(
                value=current_value,
            ).props(
                f'dense outlined hide-bottom-space step={step}'
            ).classes(field_classes)

            def _stage_literal_number(event: Any) -> None:
                self._set_pin_literal_value(
                    node,
                    pin_id=pin_id,
                    input_kind=input_kind,
                    pin_type=pin_type,
                    value=getattr(event, 'args', None),
                    notify=False,
                )
                self._mark_pending_inline_local_change()

            def _commit_literal_number(event: Any) -> None:
                self._set_pin_literal_value(
                    node,
                    pin_id=pin_id,
                    input_kind=input_kind,
                    pin_type=pin_type,
                    value=number_field.value,
                    notify=False,
                )
                self._flush_pending_inline_local_change(event)

            number_field.on(
                'update:model-value',
                _stage_literal_number,
            )
            number_field.on(
                'blur',
                _commit_literal_number,
                js_handler=(
                    '(e) => emit({'
                    'keep_pending: !!('
                    'e.relatedTarget && '
                    'e.relatedTarget.closest(".ach-node-inline-field")'
                    ')'
                    '})'
                ),
            )
            number_field.on('keydown.enter', _commit_literal_number)
            return True

        text_field = ui.input(
            value='' if current_value is None else str(current_value),
        ).props('dense outlined hide-bottom-space').classes(
            field_classes + ' ach-node-inline-field-text'
        )

        def _stage_literal_text(event: Any) -> None:
            self._set_pin_literal_value(
                node,
                pin_id=pin_id,
                input_kind=input_kind,
                pin_type=pin_type,
                value=getattr(event, 'args', None),
                notify=False,
            )
            self._mark_pending_inline_local_change()

        def _commit_literal_text(event: Any) -> None:
            self._set_pin_literal_value(
                node,
                pin_id=pin_id,
                input_kind=input_kind,
                pin_type=pin_type,
                value=text_field.value,
                notify=False,
            )
            self._flush_pending_inline_local_change(event)

        text_field.on(
            'update:model-value',
            _stage_literal_text,
        )
        text_field.on(
            'blur',
            _commit_literal_text,
            js_handler=(
                '(e) => emit({'
                'keep_pending: !!('
                'e.relatedTarget && '
                'e.relatedTarget.closest(".ach-node-inline-field")'
                ')'
                '})'
            ),
        )
        text_field.on('keydown.enter', _commit_literal_text)
        return True

    def _function_box_entry_source_id(self: Any, node: AcherionNode) -> str:
        if node.kind != 'function_box':
            return ''
        entry_node = cast(
            AcherionNode | None,
            self._function_entry_node(node.node_id),
        )
        if entry_node is None:
            return ''
        return entry_node.node_id

    def _body_input_pins(
        self: Any,
        node: AcherionNode,
    ) -> list[tuple[int, dict[str, str]]]:
        """Return input pin specs that belong in the node body."""
        top_exec = self._top_exec_input_pin(node)
        return [
            (index, pin)
            for index, pin in enumerate(self._input_pin_specs(node))
            if not _is_exec_pin(pin)
            or top_exec is None
            or index != top_exec[0]
        ]

    def _body_output_pins(
        self: Any,
        node: AcherionNode,
    ) -> list[tuple[int, dict[str, str]]]:
        """Return output pin specs that belong in the node body."""
        top_exec = self._top_exec_output_pin(node)
        return [
            (index, pin)
            for index, pin in enumerate(self._output_pin_specs(node))
            if not _is_exec_pin(pin)
            or top_exec is None
            or index != top_exec[0]
        ]

    def _top_exec_input_pin(
        self: Any,
        node: AcherionNode,
    ) -> tuple[int, dict[str, str]] | None:
        """Return the top-row exec-input pin spec, if present."""
        exec_pins = [
            (index, pin)
            for index, pin in enumerate(self._input_pin_specs(node))
            if _is_unlabeled_exec_pin(pin)
        ]
        if len(exec_pins) == 1:
            return exec_pins[0]
        return None

    def _top_exec_output_pin(
        self: Any,
        node: AcherionNode,
    ) -> tuple[int, dict[str, str]] | None:
        """Return the top-row exec-output pin spec, if present."""
        exec_pins = [
            (index, pin)
            for index, pin in enumerate(self._output_pin_specs(node))
            if _is_unlabeled_exec_pin(pin)
        ]
        if len(exec_pins) == 1:
            return exec_pins[0]
        return None

    def _full_output_source_id(
        self: Any,
        node: AcherionNode,
        pin_index: int,
    ) -> str:
        """Return the stored source id for one output pin index."""
        return f'{node.node_id}@{pin_index}'

    def _body_pin_rows(
        self: Any,
        node: AcherionNode,
    ) -> list[
        dict[
            str,
            tuple[int, dict[str, str]] | None,
        ]
    ]:
        """Return ordered body rows, pairing same-label input/output pins."""
        body_inputs = self._body_input_pins(node)
        body_outputs = self._body_output_pins(node)
        output_positions_by_key: dict[tuple[str, str], list[int]] = {}
        for output_position, (_pin_index, pin) in enumerate(body_outputs):
            match_key = _pairable_body_pin_key(pin)
            if match_key is None:
                continue
            output_positions_by_key.setdefault(match_key, []).append(
                output_position
            )

        used_output_positions: set[int] = set()
        rows: list[
            dict[str, tuple[int, dict[str, str]] | None]
        ] = []
        for input_pin in body_inputs:
            matched_output: tuple[int, dict[str, str]] | None = None
            match_key = _pairable_body_pin_key(input_pin[1])
            if match_key is not None:
                output_positions = output_positions_by_key.get(match_key, [])
                if output_positions:
                    output_position = output_positions.pop(0)
                    matched_output = body_outputs[output_position]
                    used_output_positions.add(output_position)
            rows.append({'input': input_pin, 'output': matched_output})

        for output_position, output_pin in enumerate(body_outputs):
            if output_position in used_output_positions:
                continue
            rows.append({'input': None, 'output': output_pin})
        return rows

    def _body_pin_row_count(
        self: Any,
        node: AcherionNode,
    ) -> int:
        """Return rendered body row count after inline pairing."""
        return len(self._body_pin_rows(node))

    def _body_pin_row_index(
        self: Any,
        node: AcherionNode,
        *,
        direction: str,
        pin_index: int,
    ) -> int:
        """Return body row index for one input/output pin."""
        for row_index, row in enumerate(self._body_pin_rows(node)):
            pin_entry = row['input'] if direction == 'in' else row['output']
            if pin_entry is None:
                continue
            if pin_entry[0] == pin_index:
                return row_index
        return 0

    def _render_top_exec_row(
        self: Any,
        node: AcherionNode,
    ) -> None:
        """Render unlabeled exec pins as the top row in the node body."""
        exec_input = self._top_exec_input_pin(node)
        exec_output = self._top_exec_output_pin(node)
        entry_source_id = self._function_box_entry_source_id(node)
        if (
            exec_input is None
            and exec_output is None
            and not entry_source_id
        ):
            return
        with ui.row().classes('ach-wire-row ach-exec-row'):
            with ui.row().classes('items-center gap-2'):
                if exec_input is not None:
                    pin_index, pin = exec_input
                    source_id = self._input_source_id(node, pin['pin_id'])
                    btn_cls = _pin_button_classes(
                        'ach-pin-btn ach-pin-btn-in',
                        'exec',
                    )
                    if source_id:
                        btn_cls += ' ach-pin-btn-filled'
                    ui.element('div').classes(
                        f'{btn_cls} ach-pin-anchor'
                    ).props(
                        f'data-node-id={node.node_id} '
                        'data-pin-direction=in '
                        f'data-pin-index={pin_index}'
                    ).on(
                        'click',
                        lambda _e, cur=node, pid=pin['pin_id']: (
                            self._connect_input_pin(cur, pid)
                        ),
                    )
                else:
                    ui.element('div').classes('ach-exec-row-spacer')

                if entry_source_id:
                    active = self._pending_source_node_id == entry_source_id
                    btn_cls = _pin_button_classes(
                        'ach-pin-btn ach-pin-btn-out',
                        'exec',
                    )
                    if self._has_outgoing_connection(entry_source_id):
                        btn_cls += ' ach-pin-btn-filled'
                    if active:
                        btn_cls += ' ach-pin-btn-active'
                    ui.element('div').classes(
                        f'{btn_cls} ach-pin-anchor'
                    ).props(
                        f'data-node-id={node.node_id} '
                        'data-pin-direction=out '
                        'data-pin-index=entry '
                        'title="Internal entry"'
                    ).on(
                        'click',
                        lambda _e, sid=entry_source_id: self._start_connection(sid),
                    )
            ui.space()
            if exec_output is not None:
                pin_index, _pin = exec_output
                source_id = self._full_output_source_id(node, pin_index)
                active = self._pending_source_node_id == source_id
                btn_cls = _pin_button_classes('ach-pin-btn ach-pin-btn-out', 'exec')
                if self._has_outgoing_connection(source_id):
                    btn_cls += ' ach-pin-btn-filled'
                if active:
                    btn_cls += ' ach-pin-btn-active'
                ui.element('div').classes(
                    f'{btn_cls} ach-pin-anchor'
                ).props(
                    f'data-node-id={node.node_id} '
                    'data-pin-direction=out '
                    f'data-pin-index={pin_index}'
                ).on(
                    'click',
                    lambda _e, sid=source_id: self._start_connection(sid),
                )
            else:
                ui.element('div').classes('ach-exec-row-spacer')

    def _render_input_pin_row(
        self: Any,
        node: AcherionNode,
        *,
        pin_index: int,
        pin_id: str,
        label: str,
        pin_type: str = 'any',
        editor_kind: str = '',
        optional: bool = False,
    ) -> None:
        source_id = self._input_source_id(node, pin_id)
        is_incompatible = False
        if self._pending_source_node_id is not None:
            pending_type = self._pending_output_type()
            if not _catalog_types.types_compatible(pending_type, pin_type):
                is_incompatible = True
        btn_cls = _pin_button_classes('ach-pin-btn ach-pin-btn-in', pin_type)
        if optional:
            btn_cls += ' ach-pin-btn-optional'
        if source_id:
            btn_cls += ' ach-pin-btn-filled'
        row_cls = 'ach-wire-row'
        if is_incompatible:
            row_cls += ' ach-pin-incompatible'
        if optional:
            row_cls += ' ach-wire-row-optional'
        with ui.row().classes(row_cls):
            ui.element('div').classes(
                f'{btn_cls} ach-pin-anchor'
            ).props(
                f'data-node-id={node.node_id} '
                f'data-pin-direction=in '
                f'data-pin-index={pin_index}'
            ).on(
                'click',
                lambda _e, cur=node, pid=pin_id: self._connect_input_pin(cur, pid),
            )
            ui.label(label).classes('ach-wire-label')
            self._render_input_literal_editor(
                node,
                pin_id=pin_id,
                pin_type=pin_type,
                editor_kind=editor_kind,
                source_id=source_id,
            )
            ui.space()

    def _render_output_pin_row(
        self: Any,
        node: AcherionNode,
        *,
        pin_index: int,
        label: str,
        pin_type: str = 'any',
    ) -> None:
        full_src_id = self._full_output_source_id(node, pin_index)
        active = self._pending_source_node_id == full_src_id
        btn_cls = _pin_button_classes('ach-pin-btn ach-pin-btn-out', pin_type)
        if self._has_outgoing_connection(full_src_id):
            btn_cls += ' ach-pin-btn-filled'
        if active:
            btn_cls += ' ach-pin-btn-active'
        with ui.row().classes('ach-wire-row'):
            ui.label(label).classes('ach-wire-label')
            self._render_inline_default_editor(node)
            ui.space()
            ui.element('div').classes(
                f'{btn_cls} ach-pin-anchor'
            ).props(
                f'data-node-id={node.node_id} '
                f'data-pin-direction=out '
                f'data-pin-index={pin_index}'
            ).on(
                'click',
                lambda _e, sid=full_src_id: self._start_connection(sid),
            )

    def _render_inline_pin_row(
        self: Any,
        node: AcherionNode,
        *,
        input_pin_index: int,
        input_pin_id: str,
        output_pin_index: int,
        label: str,
        pin_type: str = 'any',
        editor_kind: str = '',
        optional: bool = False,
    ) -> None:
        """Render one centered row with matching input and output pins."""
        source_id = self._input_source_id(node, input_pin_id)
        is_incompatible = False
        if self._pending_source_node_id is not None:
            pending_type = self._pending_output_type()
            if not _catalog_types.types_compatible(pending_type, pin_type):
                is_incompatible = True

        input_btn_cls = _pin_button_classes(
            'ach-pin-btn ach-pin-btn-in',
            pin_type,
        )
        if optional:
            input_btn_cls += ' ach-pin-btn-optional'
        if source_id:
            input_btn_cls += ' ach-pin-btn-filled'

        full_src_id = self._full_output_source_id(node, output_pin_index)
        active = self._pending_source_node_id == full_src_id
        output_btn_cls = _pin_button_classes(
            'ach-pin-btn ach-pin-btn-out',
            pin_type,
        )
        if self._has_outgoing_connection(full_src_id):
            output_btn_cls += ' ach-pin-btn-filled'
        if active:
            output_btn_cls += ' ach-pin-btn-active'

        row_cls = 'ach-wire-row ach-wire-row-inline'
        if is_incompatible:
            row_cls += ' ach-pin-incompatible'
        if optional:
            row_cls += ' ach-wire-row-optional'

        with ui.row().classes(row_cls):
            ui.element('div').classes(
                f'{input_btn_cls} ach-pin-anchor'
            ).props(
                f'data-node-id={node.node_id} '
                'data-pin-direction=in '
                f'data-pin-index={input_pin_index}'
            ).on(
                'click',
                lambda _e, cur=node, pid=input_pin_id: (
                    self._connect_input_pin(cur, pid)
                ),
            )
            with ui.column().classes('ach-wire-inline-center'):
                with ui.row().classes('ach-wire-inline-meta'):
                    ui.label(label).classes('ach-wire-label')
                    if not self._render_inline_default_editor(node):
                        self._render_input_literal_editor(
                            node,
                            pin_id=input_pin_id,
                            pin_type=pin_type,
                            editor_kind=editor_kind,
                            source_id=source_id,
                        )
            ui.element('div').classes(
                f'{output_btn_cls} ach-pin-anchor'
            ).props(
                f'data-node-id={node.node_id} '
                'data-pin-direction=out '
                f'data-pin-index={output_pin_index}'
            ).on(
                'click',
                lambda _e, sid=full_src_id: self._start_connection(sid),
            )

    def _render_body_pin_rows(
        self: Any,
        node: AcherionNode,
    ) -> None:
        """Render node body rows with inline matching io pairs."""
        for row in self._body_pin_rows(node):
            input_pin = row['input']
            output_pin = row['output']
            if input_pin is not None and output_pin is not None:
                input_index, input_spec = input_pin
                output_index, output_spec = output_pin
                self._render_inline_pin_row(
                    node,
                    input_pin_index=input_index,
                    input_pin_id=input_spec['pin_id'],
                    output_pin_index=output_index,
                    label=input_spec['label'],
                    pin_type=input_spec.get('type', 'any'),
                    editor_kind=input_spec.get('editor_kind', ''),
                    optional=input_spec.get('optional') == 'true',
                )
                continue
            if input_pin is not None:
                input_index, input_spec = input_pin
                self._render_input_pin_row(
                    node,
                    pin_index=input_index,
                    pin_id=input_spec['pin_id'],
                    label=input_spec['label'],
                    pin_type=input_spec.get('type', 'any'),
                    editor_kind=input_spec.get('editor_kind', ''),
                    optional=input_spec.get('optional') == 'true',
                )
                continue
            if output_pin is not None:
                output_index, output_spec = output_pin
                self._render_output_pin_row(
                    node,
                    pin_index=output_index,
                    label=output_spec['label'],
                    pin_type=output_spec.get('type', 'any'),
                )

    def _render_system_source_node(
        self: Any,
        node: AcherionNode,
    ) -> None:
        for index, pin in self._body_output_pins(node):
            self._render_output_pin_row(
                node,
                pin_index=index,
                label=pin['label'],
                pin_type=pin.get('type', 'any'),
            )

    def _render_system_sink_node(
        self: Any,
        node: AcherionNode,
        graph_index: int,  # noqa: ARG002
    ) -> None:
        self._render_body_pin_rows(node)

    def _render_function_box_node(
        self: Any,
        node: AcherionNode,
        graph_index: int,
    ) -> None:
        input_count = len(self._function_box_boundary_input_sources(node.node_id))
        output_count = len(self._function_box_boundary_output_sources(node.node_id))
        internal_nodes = self._visible_function_child_nodes(node.node_id)
        subtitle = (
            f'{_node_var_name(graph_index, node)} - '
            f'{len(internal_nodes)} nodes - '
            f'{input_count} inputs - '
            f'{output_count} outputs'
        )

        with ui.element('div').classes('ach-function-box-head'):
            ui.icon('drag_indicator').classes('ach-node-drag')
            ui.icon(_template_icon(node.kind)).classes('ach-function-box-icon')
            with ui.column().classes('gap-0 min-w-0'):
                ui.label(node.title or _template_title(node.kind)).classes(
                    'ach-function-box-title'
                )
                ui.label(subtitle).classes('ach-function-box-subtitle')
            ui.space()
            ui.button(
                icon='edit',
                on_click=lambda _e, nid=node.node_id: self._open_node_config_dialog(nid),
            ).props('flat dense round')
            ui.button(
                icon='delete',
                color='negative',
                on_click=lambda _e, nid=node.node_id: self._delete_node(nid),
            ).props('flat dense round')

        with ui.element('div').classes('ach-node-body'):
            self._render_top_exec_row(node)



"""Optional node-definition behavior helpers for compiler and embed code."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from acherion.model import AcherionNode
from acherion.registry import get_acherion_node_definition


@dataclass(frozen=True, slots=True)
class AcherionPreviewBinding:
    """Transient preview binding metadata provided by one node definition."""

    scope: str
    key: str
    input_kind: str
    file_filter: str = ''
    allow_multiple: bool = False


def _definition_field(
    definition: Any | None,
    field_name: str,
    default: Any = None,
) -> Any:
    """Return one resolved definition field or a default value."""
    if definition is None:
        return default
    reader = getattr(definition, '_definition_field', None)
    if callable(reader):
        value = reader(field_name)
        if value is not None:
            return value
    return getattr(definition, field_name, default)


def _definition_for_kind(kind: str) -> Any | None:
    """Return the registered definition for one node kind, if any."""
    return get_acherion_node_definition(str(kind or '').strip())


def _definition_for_node(node: AcherionNode) -> Any | None:
    """Return the registered definition for one node instance, if any."""
    return _definition_for_kind(node.kind)


def _bool_override(
    definition: Any | None,
    attr_name: str,
) -> bool | None:
    """Return one optional bool override from a definition."""
    if definition is None or not hasattr(definition, attr_name):
        return None
    return bool(getattr(definition, attr_name))


def _string_override(
    definition: Any | None,
    attr_name: str,
) -> str:
    """Return one optional string override from a definition."""
    if definition is None:
        return ''
    value = getattr(definition, attr_name, '')
    if callable(value):
        try:
            value = value()
        except TypeError:
            return ''
    return str(value or '').strip()


def is_ui_node_kind(kind: str) -> bool:
    """Return True when one node kind is registered as a UI node."""
    definition = _definition_for_kind(kind)
    category = str(_definition_field(definition, 'category', '') or '')
    return category == 'ui'


def compiler_is_backward_getter(node: AcherionNode) -> bool:
    """Return True when one node should seed dependency aliases backwards."""
    definition = _definition_for_node(node)
    override = _bool_override(definition, 'compiler_backward_getter')
    if override is not None:
        return override
    return is_ui_node_kind(node.kind)


def compiler_is_serial_root(node: AcherionNode) -> bool:
    """Return True when one node can start an implicit serial chain."""
    definition = _definition_for_node(node)
    override = _bool_override(definition, 'compiler_serial_root')
    if override is not None:
        return override
    return is_ui_node_kind(node.kind)


def compiler_can_reemit(node: AcherionNode) -> bool:
    """Return True when one node can be emitted again deeper in scope."""
    definition = _definition_for_node(node)
    override = _bool_override(definition, 'compiler_reemittable')
    if override is not None:
        return override
    flavor = str(_definition_field(definition, 'flavor', '') or '')
    return flavor == 'pure' or is_ui_node_kind(node.kind)


def compiler_has_zero_data_exec_output(node: AcherionNode) -> bool:
    """Return True when exec output occupies pin index zero."""
    definition = _definition_for_node(node)
    override = _bool_override(definition, 'compiler_zero_data_output')
    if override is not None:
        return override
    return bool(_definition_field(definition, 'exec_out', False)) and not bool(
        _definition_field(definition, 'producer', False)
    )


def function_box_unsupported(node: AcherionNode) -> bool:
    """Return True when one node should not compile inside a Function Box."""
    definition = _definition_for_node(node)
    override = _bool_override(definition, 'function_box_unsupported')
    if override is not None:
        return override
    category = str(_definition_field(definition, 'category', '') or '')
    return category == 'effect' or bool(
        _definition_field(definition, 'system_sink', False)
    )


def dependency_expr_for_node(
    node: AcherionNode,
    *,
    source_expr: str | None = None,
    owner_name: str | None = 'self',
) -> str | None:
    """Return dependency alias expression for one node, when supported."""
    definition = _definition_for_node(node)
    method = getattr(definition, 'dependency_expr', None)
    if not callable(method):
        return None
    result = method(
        node,
        source_expr=source_expr,
        owner_name=owner_name,
    )
    if result is None:
        return None
    return str(result)


def emit_static_node(
    *,
    state: Any,
    node: AcherionNode,
    params: dict[str, Any],
    indent: str,
    var_name: str,
    owner_name: str | None = 'self',
) -> bool:
    """Delegate one static node emission to its registered definition."""
    definition = _definition_for_node(node)
    method = getattr(definition, 'emit_static_node', None)
    if not callable(method):
        return False
    return bool(
        method(
            node,
            state=state,
            params=params,
            indent=indent,
            var_name=var_name,
            owner_name=owner_name,
        )
    )


def inline_default_editor_spec_for_node(
    node: AcherionNode,
) -> tuple[str, str, Any] | None:
    """Return compact inline editor metadata for one node, if any."""
    definition = _definition_for_node(node)
    method = getattr(definition, 'inline_default_editor_spec', None)
    if not callable(method):
        return None
    result = method(node)
    if result is None:
        return None
    return result


def render_inline_controls_for_node(
    owner: Any,
    node: AcherionNode,
) -> bool:
    """Delegate compact inline node controls to one registered definition."""
    definition = _definition_for_node(node)
    method = getattr(definition, 'render_inline_controls', None)
    if not callable(method):
        return False
    return bool(method(owner, node))


def preview_binding_for_node(
    node: AcherionNode,
) -> AcherionPreviewBinding | None:
    """Return transient preview input binding for one node, if any."""
    definition = _definition_for_node(node)
    method = getattr(definition, 'preview_binding', None)
    if not callable(method):
        return None
    result = method(node)
    if result is None:
        return None
    return result


def preview_result_reference_for_node(
    node: AcherionNode,
) -> str:
    """Return transient preview result reference for one node, if any."""
    definition = _definition_for_node(node)
    method = getattr(definition, 'preview_result_reference', None)
    if not callable(method):
        return ''
    return str(method(node) or '').strip()


def node_type_summary(kind: str) -> str:
    """Return human-readable type summary provided by one definition."""
    return _string_override(_definition_for_kind(kind), 'type_summary')


def output_pin_label(kind: str) -> str:
    """Return output-pin label override provided by one definition."""
    return _string_override(_definition_for_kind(kind), 'output_pin_label')


def sink_pin_label(kind: str) -> str:
    """Return sink-pin label override provided by one definition."""
    return _string_override(_definition_for_kind(kind), 'sink_pin_label')


def sink_input_type_tag(kind: str) -> str:
    """Return sink input type override provided by one definition."""
    return _string_override(_definition_for_kind(kind), 'sink_input_type_tag')


__all__ = [
    'AcherionPreviewBinding',
    'compiler_can_reemit',
    'compiler_has_zero_data_exec_output',
    'compiler_is_backward_getter',
    'compiler_is_serial_root',
    'dependency_expr_for_node',
    'emit_static_node',
    'function_box_unsupported',
    'inline_default_editor_spec_for_node',
    'is_ui_node_kind',
    'node_type_summary',
    'output_pin_label',
    'render_inline_controls_for_node',
    'preview_binding_for_node',
    'preview_result_reference_for_node',
    'sink_input_type_tag',
    'sink_pin_label',
]